import argparse

from ptpython.repl import embed


def run():
    from nekoyume.app import app

    parser = argparse.ArgumentParser(description='Nekoyume shell')
    args = parser.parse_args()

    app.app_context().push()
    embed(globals(), locals())


if __name__ == '__main__':
    run()



class Item():
    ticker_name = 'XXX'
    pass


class Combined():
    recipe = {}
    dice = '1d1'


class Weapon(Item):
    @classmethod
    def attack_modifier(cls, avatar, monster):
        raise NotImplementedError


class Armor(Item):
    @classmethod
    def armor_modifier(cls, avatar, monster):
        raise NotImplementedError


class Food(Item):
    def execute(self, avatar):
        raise NotImplementedError


class HP10Food(Food):
    def execute(self, avatar):
        avatar.hp = min(avatar.max_hp, avatar.hp + 10)
        return avatar, {
            'type': 'item_use',
            'item': self.ticker_name,
            'status_change': 'HP +10'
        }


class HPMaxFood(Food):
    def execute(self, avatar):
        avatar.hp = avatar.max_hp
        return avatar, {
            'type': 'item_use',
            'item': self.ticker_name,
            'status_change': 'HP MAX'
        }


class MeatPlatter(Combined, HP10Food):
    ticker_name = 'MPLT'
    recipe = {'MEAT', 'MEAT', 'MEAT'}


class Oyakodong(Combined, HP10Food):
    ticker_name = 'OYKD'
    recipe = {'RICE', 'EGGS', 'CHKN'}


class Carbonara(Combined, HP10Food):
    ticker_name = 'CBNR'
    recipe = {'WHET', 'EGGS', 'MEAT'}


class Steakdong(Combined, HP10Food):
    ticker_name = 'STKD'
    recipe = {'RICE', 'RKST', 'MEAT'}


class ChickenRice(Combined, HP10Food):
    ticker_name = 'CHKR'
    recipe = {'RICE', 'RKST', 'CHKN'}


class Steak(Combined, HP10Food):
    ticker_name = 'STEK'
    recipe = {'MEAT', 'RKST', 'OLIV'}


class SteakCarbonara(Combined, HPMaxFood):
    ticker_name = 'STCB'
    recipe = {'STEK', 'WHET', 'EGGS'}


class FriedChicken(Combined, HP10Food):
    ticker_name = 'FCHK'
    recipe = {'CHKN', 'RKST', 'OLIV'}


class LongSword(Weapon):
    ticker_name = 'LSWD'

    @classmethod
    def attack_modifier(cls, avatar, monster):
        return 1


class FlameSword(Combined, Weapon):
    ticker_name = 'FSWD'
    recipe = {'LSWD', 'FLNT', 'OLIV'}

    @classmethod
    def attack_modifier(cls, avatar, monster):
        if 'Wooden' in monster['special']:
            return 2
        else:
            return 1


class FlameSword1(Combined, Weapon):
    ticker_name = 'FSW1'
    recipe = {'FSWD', 'FSWD', 'FSWD'}

    @classmethod
    def attack_modifier(cls, avatar, monster):
        if 'Wooden' in monster['special']:
            return 4
        else:
            return 2


class FlameSword2(Combined, Weapon):
    ticker_name = 'FSW2'
    recipe = {'FSW1', 'FSW1', 'FSW1'}

    @classmethod
    def attack_modifier(cls, avatar, monster):
        if 'Wooden' in monster['special']:
            return 6
        else:
            return 3


class FlameSword3(Combined, Weapon):
    ticker_name = 'FSW3'
    recipe = {'FSW2', 'FSW2', 'FSW2'}

    @classmethod
    def attack_modifier(cls, avatar, monster):
        if 'Wooden' in monster['special']:
            return 8
        else:
            return 4


class LeatherMail(Armor):
    ticker_name = 'LMIL'

    @classmethod
    def armor_modifier(cls, avatar, monster):
        return 1


def get_related_items(cls) -> set:
    subclasses = set()
    pool = [cls]
    while pool:
        parent = pool.pop()
        for child in parent.__subclasses__():
            if child not in subclasses:
                if child.ticker_name != 'XXX':
                    subclasses.add(child)
                pool.append(child)
    return subclasses



#!/usr/bin/env python3.6
"""Implement Hashcash version 1 protocol in Python 3.6
+---------------------------------------------------------------------------+
| Written by David Mertz; Modified by Jc Kim; released to the Public Domain |
+---------------------------------------------------------------------------+

Double spend database not implemented in this module, but stub
for callbacks is provided in the 'check()' function

The function 'check()' will validate 'generalized hashcash' tokens
generically.  Future protocol version are treated as generalized tokens
(should a future version be published w/o this module being correspondingly
updated).

A 'generalized hashcash' is implemented in the '_mint()' function, with the
public function 'mint()' providing a wrapper for actual hashcash protocol.
The generalized form simply finds a suffix that creates zero bits in the
hash of the string concatenating 'challenge' and 'suffix' without specifying
any particular fields or delimiters in 'challenge'.  E.g., you might get:

    >>> from nekoyume.hashcash import _mint
    >>> _mint('foo', bits=16)
    '1369a'
    >>> from hashlib import sha256
    >>> sha256(b'foo1369a').hexdigest()
    '0000ac007db1be2dcb5e1fb94685c64976c8485931a44b8546d3034ad853555f'
    >>> sha256(b'1:16:180130:foo::QiTuo+Vq:19556').hexdigest()
    '00002d016664c69bda0fe9730f058fa8b5ef508fc8078253e07948c769d4642a'

Notice that '_mint()' behaves deterministically, finding the same suffix
every time it is passed the same arguments.  'mint()' incorporates a random
salt in stamps (as per the hashcash v.1 protocol).
"""
from math import ceil, floor
from hashlib import sha256


tries = [0]                 # Count hashes performed for benchmark


def _mint(challenge, bits):
    """Answer a 'generalized hashcash' challenge'

    Hashcash requires stamps of form 'ver:bits:date:res:ext:rand:counter'
    This internal function accepts a generalized prefix 'challenge',
    and returns only a suffix that produces the requested SHA leading zeros.

    NOTE: Number of requested bits is rounded up to the nearest multiple of 4
    """
    counter = 0
    hex_digits = int(ceil(bits/4.))
    zeros = '0'*hex_digits
    while 1:
        digest = sha256(str.encode(challenge+hex(counter)[2:])).hexdigest()
        if digest[:hex_digits] == zeros:
            tries[0] = counter
            return hex(counter)[2:]
        counter += 1


def check(stamp, resource=None, bits=None,
          check_expiration=None, ds_callback=None):
    if type(bits) is not int:
        return True
    elif resource is not None and stamp.find(resource) < 0:
        return False
    else:
        hex_digits = int(floor(bits/4))
        return sha256(str.encode(stamp)).hexdigest().startswith(
            '0'*hex_digits)



"""
Models
======

`models.py` contains every relations regarding nekoyume blockchain and
game moves.
"""

import datetime
from hashlib import sha256 as h
import os

from bencode import bencode
from bitcoin import base58
from flask_cache import Cache
from flask_sqlalchemy import SQLAlchemy
import requests
import seccure
from sqlalchemy import or_
from sqlalchemy.ext.associationproxy import association_proxy
from sqlalchemy.orm.collections import attribute_mapped_collection
import tablib

from nekoyume.exc import (InvalidBlockError,
                          InvalidMoveError,
                          InvalidNameError,
                          OutOfRandomError)
from nekoyume.items import (Armor,
                            Combined,
                            get_related_items,
                            Item,
                            Weapon,
                            Food)
from nekoyume import hashcash


db = SQLAlchemy()
cache = Cache()


def get_my_public_url():
    if 'PUBLIC_URL' in os.environ:
        return os.environ['PUBLIC_URL']
    try:
        if os.environ.get('PORT', '80') != '80':
            port = ':' + os.environ.get('PORT', '80')
        else:
            port = ''
        ip = requests.get('http://ip.42.pl/raw').text
        has_public_address = requests.get(
            f'http://{ip}{port}/ping'
        ).text == 'pong'
    except requests.exceptions.ConnectionError:
        return None
    if has_public_address:
        return f'http://{ip}{port}'
    else:
        return None


class Node(db.Model):
    """This object contains node information you know."""

    #: URL of node
    url = db.Column(db.String, primary_key=True)
    #: last connected datetime of the node
    last_connected_at = db.Column(db.DateTime, nullable=False, index=True)

    get_nodes_endpoint = '/nodes'
    post_node_endpoint = '/nodes'
    get_blocks_endpoint = '/blocks'
    post_block_endpoint = '/blocks'
    post_move_endpoint = '/moves'

    @classmethod
    def get(cls, url, session=db.session):
        node = Node.query.filter_by(url=url).first()
        if node:
            return node
        elif requests.get(f'{url}/ping').text == 'pong':
            node = Node(url=url, last_connected_at=datetime.datetime.utcnow())
            if session:
                session.add(node)
                session.commit()
            return node
        else:
            return None

    @classmethod
    def update(cls, node=None):
        """
        Update recent node list by scrapping other nodes' information.
        """
        if not node or not node.url:
            recent_nodes = Node.query.filter(
                Node.last_connected_at >= datetime.datetime.utcnow() -
                datetime.timedelta(60 * 3)
            ).order_by(Node.last_connected_at.desc()).limit(2500)
            if Node.query.count() == 0:
                recent_nodes = [cls(url='http://seed.nekoyu.me')]
        else:
            recent_nodes = [node]

        for n in recent_nodes:
            try:
                response = requests.get(f"{n.url}{Node.get_nodes_endpoint}")
                for url in response.json()['nodes']:
                    Node.get(url)
                else:
                    continue
            except requests.exceptions.ConnectionError:
                continue
            break

        db.session.commit()

    def ping(self):
        try:
            result = requests.get(f'{self.url}/ping').text == 'pong'
            if result:
                self.last_connected_at = datetime.datetime.utcnow()
            return result
        except requests.exceptions.ConnectionError:
            return False

    @classmethod
    def broadcast(cls,
                  endpoint: str,
                  serialized_obj: dict,
                  sent_node=None,
                  my_node=None,
                  session=db.session) -> bool:
        """
        It broadcast `serialized_obj` to every nodes you know.

        :param        endpoint: endpoint of node to broadcast
        :param  serialized_obj: object that will be broadcasted.
        :param       sent_node: sent :class:`nekoyume.models.Node`.
                                this node ignore sent node.
        :param         my_node: my :class:`nekoyume.models.Node`.
                                received node ignore my node when they
                                broadcast received object.
        """

        for node in session.query(cls):
            if sent_node and sent_node.url == node.url:
                continue
            try:
                if my_node:
                    serialized_obj['sent_node'] = my_node.url
                requests.post(node.url + endpoint, json=serialized_obj)
                node.last_connected_at = datetime.datetime.utcnow()
                session.add(node)
            except requests.exceptions.ConnectionError:
                continue

        session.commit()
        return True


class Block(db.Model):
    """This object contains block information."""

    __tablename__ = 'block'
    #: block id
    id = db.Column(db.Integer, primary_key=True)
    #: current block's hash
    hash = db.Column(db.String, nullable=False, index=True, unique=True)
    #: previous block's hash
    prev_hash = db.Column(db.String, index=True)
    #: block creator's address
    creator = db.Column(db.String, nullable=False, index=True)
    #: hash of every linked move's ordered hash list
    root_hash = db.Column(db.String, nullable=False)
    #: suffix for hashcash
    suffix = db.Column(db.String, nullable=False)
    #: difficulty of hashcash
    difficulty = db.Column(db.Integer, nullable=False)
    #: block creation datetime
    created_at = db.Column(db.DateTime, nullable=False,
                           default=datetime.datetime.utcnow())
    size_limit = 10000

    @property
    def valid(self) -> bool:
        """Check if this object is valid or not"""
        stamp = self.serialize().decode('utf-8') + self.suffix
        valid = (self.hash == h(str.encode(stamp)).hexdigest())
        valid = valid and hashcash.check(stamp, self.suffix, self.difficulty)

        valid = valid and (
            len(self.serialize(True, True, True, True)) <= Block.size_limit
        )

        if self.id > 1:
            prev_block = Block.query.get(self.id - 1)
            valid = valid and self.prev_hash == prev_block.hash

            difficulty = prev_block.difficulty
            difficulty_check_block = Block.query.get(
                max(1, self.id - 10)
            )
            avg_timedelta = (
                (self.created_at - difficulty_check_block.created_at) /
                (self.id - difficulty_check_block.id)
            )
            if avg_timedelta <= datetime.timedelta(0, 5):
                valid = valid and self.difficulty == max(0, difficulty + 1)
            elif avg_timedelta > datetime.timedelta(0, 15):
                valid = valid and self.difficulty == max(0, difficulty - 1)
            else:
                valid = valid and self.difficulty == difficulty
        else:
            valid = valid and self.prev_hash is None
            valid = valid and self.difficulty == 0

        valid = valid and self.root_hash == h(
            ''.join(sorted((m.id for m in self.moves))).encode('utf-8')
        ).hexdigest()

        for move in self.moves:
            valid = valid and move.valid
        return valid

    def serialize(self,
                  use_bencode: bool=True,
                  include_suffix: bool=False,
                  include_moves: bool=False,
                  include_hash: bool=False):
        """
        This function serialize block.

        :param    use_bencode: check if you want to encode using bencode.
        :param include_suffix: check if you want to include suffix.
        :param  include_moves: check if you want to include linked moves.
        :param   include_hash: check if you want to include block hash.
        """
        serialized = dict(
            id=self.id,
            creator=self.creator,
            prev_hash=self.prev_hash,
            difficulty=self.difficulty,
            root_hash=self.root_hash,
            created_at=str(self.created_at),
        )
        if include_suffix:
            serialized['suffix'] = self.suffix

        if include_moves:
            serialized['moves'] = [m.serialize(
                use_bencode=False,
                include_signature=True,
                include_id=True,
            ) for m in self.moves]

        if include_hash:
            serialized['hash'] = self.hash

        if use_bencode:
            if self.prev_hash is None:
                del serialized['prev_hash']
            serialized = bencode(serialized)
        return serialized

    def broadcast(self,
                  sent_node: bool=None,
                  my_node: bool=None,
                  session=db.session,
                  click=None) -> bool:
        """
        It broadcast this block to every nodes you know.

       :param       sent_node: sent :class:`nekoyume.models.Node`.
                               this node ignore sent node.
       :param         my_node: my :class:`nekoyume.models.Node`.
                               received node ignore my node when they
                               broadcast received object.
        """
        return Node.broadcast(Node.post_block_endpoint,
                              self.serialize(False, True, True, True),
                              sent_node, my_node, session)

    @classmethod
    def sync(cls, node: Node=None, session=db.session, click=None) -> bool:
        """
        Sync blockchain with other node.

        :param node: sync target :class:`nekoyume.models.Node`.
        """
        if not node:
            nodes = Node.query.filter(
                Node.last_connected_at >= datetime.datetime.utcnow() -
                datetime.timedelta(60 * 3)
            ).order_by(Node.last_connected_at.desc())
        else:
            nodes = [node]

        if not nodes:
            return False

        node_last_block = None
        for node in nodes:
            try:
                response = requests.get(
                    f"{node.url}{Node.get_blocks_endpoint}/last"
                )
                node_last_block = response.json()['block']
                break
            except requests.exceptions.ConnectionError:
                continue

        last_block = session.query(Block).order_by(Block.id.desc()).first()

        if not node_last_block:
            return True

        #: If my chain is the longest one, we don't need to do anything.
        if last_block and last_block.id >= node_last_block['id']:
            return True

        def find_branch_point(value: int, high: int):
            mid = int((value + high) / 2)
            response = requests.get((f"{node.url}{Node.get_blocks_endpoint}/"
                                     f"{mid}"))
            block = session.query(Block).get(mid)
            if value > high:
                return 0
            if (response.json()['block'] and block and
               block.hash == response.json()['block']['hash']):
                if value == mid:
                        return value
                return find_branch_point(mid, high)
            else:
                return find_branch_point(value, mid - 1)

        if last_block:
            # TODO: Very hard to understand. fix this easily.
            if find_branch_point(last_block.id,
                                 last_block.id) == last_block.id:
                branch_point = last_block.id
            else:
                branch_point = find_branch_point(0, last_block.id)
        else:
            branch_point = 0

        for block in session.query(Block).filter(Block.id > branch_point):
            for move in block.moves:
                move.block_id = None
            session.delete(block)

        from_ = branch_point + 1
        limit = 1000
        while True:
            if click:
                click.echo(f'Syncing blocks...(from: {from_})')
            response = requests.get(f"{node.url}{Node.get_blocks_endpoint}",
                                    params={'from': from_,
                                            'to': from_ + limit - 1})
            if len(response.json()['blocks']) == 0:
                break
            for new_block in response.json()['blocks']:
                block = Block()
                block.id = new_block['id']
                block.creator = new_block['creator']
                block.created_at = datetime.datetime.strptime(
                    new_block['created_at'], '%Y-%m-%d %H:%M:%S.%f')
                block.prev_hash = new_block['prev_hash']
                block.hash = new_block['hash']
                block.difficulty = new_block['difficulty']
                block.suffix = new_block['suffix']
                block.root_hash = new_block['root_hash']

                for new_move in new_block['moves']:
                    move = session.query(Move).get(new_move['id'])
                    if not move:
                        move = Move(
                            id=new_move['id'],
                            user=new_move['user'],
                            name=new_move['name'],
                            signature=new_move['signature'],
                            tax=new_move['tax'],
                            details=new_move['details'],
                            created_at=datetime.datetime.strptime(
                                new_move['created_at'],
                                '%Y-%m-%d %H:%M:%S.%f'),
                            block_id=block.id,
                        )
                    if not move.valid:
                        session.rollback()
                        raise InvalidMoveError
                    block.moves.append(move)
                    session.add(move)

                if not block.valid:
                    session.rollback()
                    raise InvalidBlockError
                session.add(block)
            if len(response.json()['blocks']) < 1000:
                break
            from_ += limit
            db.session.commit()
        db.session.commit()
        return True


def get_address(public_key: str) -> str:
    """Get address based on the public key"""
    return base58.encode(public_key)


class Move(db.Model):
    """This object contain general move information."""
    __tablename__ = 'move'
    #: move's hash
    id = db.Column(db.String, primary_key=True)
    #: move's block id. if the move isn't confirmed yet, this will be null
    block_id = db.Column(db.Integer, db.ForeignKey('block.id'),
                         nullable=True, index=True)
    #: move's block
    block = db.relationship('Block', uselist=False, backref='moves')
    #: move's owner
    user = db.Column(db.String, nullable=False, index=True)
    #: move's signature
    signature = db.Column(db.String, nullable=False)
    #: move name
    name = db.Column(db.String, nullable=False, index=True)
    #: move details. it contains parameters of move
    details = association_proxy(
        'move_details', 'value',
        creator=lambda k, v: MoveDetail(key=k, value=v)
    )
    #: move tax (not implemented yet)
    tax = db.Column(db.BigInteger, default=0, nullable=False)
    #: move creation datetime.
    created_at = db.Column(db.DateTime, nullable=False,
                           default=datetime.datetime.utcnow())

    __mapper_args__ = {
        'polymorphic_identity': 'move',
        'polymorphic_on': name,
    }

    @property
    def valid(self):
        """Check if this object is valid or not"""
        if not self.signature or self.signature.find(' ') < 0:
            return False

        public_key = self.signature.split(' ')[1]
        valid = True

        valid = valid and seccure.verify(
            self.serialize(include_signature=False),
            self.signature.split(' ')[0],
            public_key,
        )
        valid = valid and (
            self.user == get_address(public_key.encode('utf-8'))
        )

        valid = valid and (self.id == self.hash)

        return valid

    @property
    def confirmed(self):
        """Check if this object is confirmed or not"""
        return self.block and self.block.hash is not None

    def serialize(self,
                  use_bencode=True,
                  include_signature=False,
                  include_id=False,
                  include_block=False):
        """
        This function serialize block.

        :param       use_bencode: check if you want to encode using bencode.
        :param include_signature: check if you want to include signature.
        :param        include_id: check if you want to include linked moves.
        :param     include_block: check if you want to include block.
        """
        serialized = dict(
            user=self.user,
            name=self.name,
            details={k: str(v) for k, v in dict(self.details).items()},
            tax=self.tax,
            created_at=str(self.created_at),
        )
        if include_signature:
            serialized['signature'] = self.signature
        if include_id:
            serialized['id'] = self.id
        if include_block:
            if self.block:
                serialized['block'] = self.block.serialize(False)
            else:
                serialized['block'] = None
        if use_bencode:
            serialized = bencode(serialized)
        return serialized

    def broadcast(self, sent_node=None, my_node=None, session=db.session):
        """
        It broadcast this move to every nodes you know.

       :param       sent_node: sent :class:`nekoyume.models.Node`.
                               this node ignore sent node.
       :param         my_node: my :class:`nekoyume.models.Node`.
                               received node ignore my node when they
                               broadcast received object.
        """
        Node.broadcast(Node.post_move_endpoint,
                       self.serialize(False, True, True),
                       sent_node, my_node, session)

    @property
    def hash(self) -> str:
        """ Get move hash """
        return h(self.serialize(include_signature=True)).hexdigest()

    def get_randoms(self) -> list:
        """ get random numbers by :doc:`Hash random <white_paper>` """
        if not (self.block and self.block.hash and self.id):
            return []
        result = [ord(a) ^ ord(b) for a, b in zip(self.block.hash, self.id)]
        result = result[int(self.block.difficulty / 4):]
        return result

    def roll(self, randoms: list, dice: str, combine=True):
        """
        Roll dices based on given randoms

            >>> from nekoyume.models import Move
            >>> move = Move()
            >>> move.roll([1, 7, 3], '2d6')
            6

        :params randoms: random numbers from
                         :func:`nekoyume.models.Move.get_randoms`
        :params    dice: dice to roll (e.g. 2d6)
        :params combine: return combined result or not if rolling it multiple.
        """
        result = []
        if dice.find('+') > 0:
            dice, plus = dice.split('+')
            plus = int(plus)
        else:
            plus = 0
        cnt, dice_type = (int(i) for i in dice.split('d'))
        for i in range(cnt):
            try:
                result.append(randoms.pop() % dice_type + 1)
            except IndexError:
                raise OutOfRandomError
        if combine:
            return sum(result) + plus
        else:
            return result


class MoveDetail(db.Model):
    """ This object contains move's key/value information. """

    #: move id
    move_id = db.Column(db.String,  db.ForeignKey('move.id'),
                        nullable=True, primary_key=True)
    move = db.relationship(Move, backref=db.backref(
        'move_details',
        collection_class=attribute_mapped_collection("key"),
        cascade="all, delete-orphan"
    ))
    #: MoveDetail's key
    key = db.Column(db.String, nullable=False, primary_key=True)
    #: MoveDetail's value
    value = db.Column(db.String, nullable=False, index=True)


class HackAndSlash(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'hack_and_slash',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)
        dirname = os.path.dirname(__file__)
        filename = os.path.join(dirname, 'data/monsters.csv')
        monsters = tablib.Dataset().load(
            open(filename).read()
        ).dict
        randoms = self.get_randoms()
        monster = monsters[randoms.pop() % len(monsters)]
        battle_status = []

        for key in ('hp', 'piercing', 'armor'):
            monster[key] = int(monster[key])

        def get_item(ticker_name):
            items = get_related_items(Item)
            for item in items:
                if item.ticker_name == ticker_name:
                    return item
            return None

        weapon = armor = food = None
        if 'weapon' in self.details:
            weapon = get_item(self.details['weapon'])
        if 'armor' in self.details:
            armor = get_item(self.details['armor'])
        if 'food' in self.details:
            food = get_item(self.details['food'])

        while True:
            try:
                if (avatar.hp <= avatar.max_hp * 0.5
                   and food and food.ticker_name in avatar.items
                   and avatar.items[food.ticker_name] > 0):
                    avatar, status = food().execute(avatar)
                    battle_status.append(status)
                    avatar.items[food.ticker_name] -= 1
                    food = None

                if (avatar.hp <= avatar.max_hp * 0.2
                   and 'BNDG' in avatar.items and avatar.items['BNDG'] > 0):
                    rolled = self.roll(randoms, '2d6')
                    if rolled >= 7:
                        avatar.hp += 4
                        avatar.items['BNDG'] -= 1
                        battle_status.append({
                            'type': 'item_use',
                            'item': 'BNDG',
                            'status_change': 'HP +4'
                        })
                    else:
                        avatar.items['BNDG'] -= 1
                        battle_status.append({
                            'type': 'item_use_fail',
                            'item': 'BNDG',
                            'status_change': ''
                        })

                rolled = (self.roll(randoms, '2d6')
                          + avatar.modifier('strength'))
                if rolled >= 7:
                    damage = max(
                        self.roll(randoms, avatar.damage) - monster['armor'], 0
                    )
                    if weapon:
                        damage += weapon.attack_modifier(avatar, monster)
                    battle_status.append({
                        'type': 'attack_monster',
                        'damage': damage,
                        'monster': monster.copy(),
                    })
                    monster['hp'] = monster['hp'] - damage

                elif rolled in (2, 3, 4, 5, 6, 7, 8, 9):
                    monster_damage = self.roll(randoms, monster['damage'])
                    if armor:
                        monster_damage -= armor.armor_modifier(avatar, monster)
                    battle_status.append({
                        'type': 'attacked_by_monster',
                        'damage': monster_damage,
                        'monster': monster.copy(),
                    })
                    avatar.hp -= monster_damage
                    if rolled <= 6:
                        battle_status.append({
                            'type': 'get_xp',
                        })
                        avatar.xp += 1

                if monster['hp'] <= 0:
                    battle_status.append({
                        'type': 'kill_monster',
                        'monster': monster.copy(),
                    })
                    reward_code = self.roll(randoms, '1d10')
                    if len(monster[f'reward{reward_code}']):
                        avatar.get_item(monster[f'reward{reward_code}'])
                        battle_status.append({
                            'type': 'get_item',
                            'item': monster[f'reward{reward_code}'],
                        })
                    return (avatar, dict(
                        type='hack_and_slash',
                        result='win',
                        battle_status=battle_status,
                    ))

                if avatar.hp <= 0:
                    battle_status.append({
                        'type': 'killed_by_monster',
                        'monster': monster.copy(),
                    })
                    return (avatar, dict(
                        type='hack_and_slash',
                        result='lose',
                        battle_status=battle_status,
                    ))

            except OutOfRandomError:
                battle_status.append({
                    'type': 'run',
                    'monster': monster.copy(),
                })
                return (avatar, dict(
                    type='hack_and_slash',
                    result='finish',
                    battle_status=battle_status,
                ))


class Sleep(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'sleep',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)
        avatar.hp = avatar.max_hp
        return avatar, dict(
            type='sleep',
            result='success',
        )


class CreateNovice(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'create_novice',
    }

    def execute(self, avatar=None):
        if avatar:
            #: Keep the information that should not be removed.
            gold = avatar.items['GOLD']
        else:
            gold = 0
        avatar = Novice()

        avatar.strength = int(self.details['strength'])
        avatar.dexterity = int(self.details['dexterity'])
        avatar.constitution = int(self.details['constitution'])
        avatar.intelligence = int(self.details['intelligence'])
        avatar.wisdom = int(self.details['wisdom'])
        avatar.charisma = int(self.details['charisma'])

        if (avatar.strength + avatar.dexterity + avatar.constitution +
           avatar.intelligence + avatar.wisdom + avatar.charisma) > 64:
            avatar.strength = 9
            avatar.dexterity = 9
            avatar.constitution = 9
            avatar.intelligence = 9
            avatar.wisdom = 9
            avatar.charisma = 9

        if 'name' in self.details:
            avatar.name = self.details['name']
        else:
            avatar.name = self.user[:6]

        if 'gravatar_hash' in self.details:
            avatar.gravatar_hash = self.details['gravatar_hash']
        else:
            avatar.gravatar_hash = 'HASH'

        avatar.user = self.user
        avatar.current_block = self.block
        avatar.hp = avatar.max_hp
        avatar.xp = 0
        avatar.lv = 1
        avatar.items = dict(
            GOLD=gold
        )

        return (avatar, dict(
            type='create_novice',
            result='success',
        ))


class LevelUp(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'level_up',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)
        if avatar.xp < avatar.lv + 7:
            return avatar, dict(
                type='level_up',
                result='failed',
                message="You don't have enough xp.",
            )

        avatar.xp -= avatar.lv + 7
        avatar.lv += 1
        setattr(avatar, self.details['new_status'],
                getattr(avatar, self.details['new_status']) + 1)
        if self.details['new_status'] == 'constitution':
            avatar.hp += 1
        return avatar, dict(
            type='level_up',
            result='success',
        )


class Say(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'say',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)

        return avatar, dict(
            type='say',
            message=self.details['content'],
        )


class Send(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'send',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)

        if (self.details['item_name'] not in avatar.items or
           avatar.items[self.details['item_name']]
           - int(self.details['amount']) < 0):
            return avatar, dict(
                type='send',
                result='fail',
                message="You don't have enough items to send."
            )

        avatar.items[self.details['item_name']] -= int(self.details['amount'])
        return avatar, dict(
            type='send',
            result='success',
        )

    def receive(self, receiver=None):
        if not receiver:
            receiver = Avatar.get(self.details['receiver'], self.block_id - 1)

        for i in range(int(self.details['amount'])):
            receiver.get_item(self.details['item_name'])

        return receiver, dict(
            type='receive',
            result='success',
        )


class Combine(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'combine',
    }

    def execute(self, avatar=None):
        if not avatar:
            avatar = Avatar.get(self.user, self.block_id - 1)
        if avatar.items['GOLD'] <= 0:
            return avatar, dict(
                type='combine',
                result='failure',
                reason='insufficient_gold'
            )
        for i in ('item1', 'item2', 'item3'):
            if (self.details[i] not in avatar.items or
               avatar.items[self.details[i]] <= 0):
                return avatar, dict(
                    type='combine',
                    result='failure',
                    reason='insufficient_item'
                )
        randoms = self.get_randoms()
        recipes = {scls.ticker_name: scls.recipe
                   for scls in Combined.__subclasses__()}
        dices = {scls.ticker_name: scls.dice
                 for scls in Combined.__subclasses__()}
        for result, recipe in recipes.items():
            if recipe == {self.details['item1'],
                          self.details['item2'],
                          self.details['item3']}:
                avatar.items[self.details['item1']] -= 1
                avatar.items[self.details['item2']] -= 1
                avatar.items[self.details['item3']] -= 1
                avatar.items['GOLD'] -= 1
                if self.roll(randoms, dices[result]) == 1:
                    avatar.get_item(result)
                    return avatar, dict(
                        type='combine',
                        result='success',
                        result_item=result,
                    )
                else:
                    return avatar, dict(
                        type='combine',
                        result='failure',
                        reason='bad_luck'
                    )

        return avatar, dict(
            type='combine',
            result='failure',
            reason='no_combination'
        )


class Sell(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'sell',
    }


class Buy(Move):
    __mapper_args__ = {
        'polymorphic_identity': 'buy',
    }


class User():
    """ It contains user's keys and avatar information. """

    def __init__(self, private_key, session=db.session):
        self.private_key = private_key
        self.public_key = str(seccure.passphrase_to_pubkey(
            self.private_key.encode('utf-8')
        ))
        self.session = session

    @property
    def address(self) -> str:
        """ It returns address of the user. """
        return get_address(self.public_key.encode('utf-8'))

    def sign(self, move: Move):
        """ put signature into the given move using the user's private key. """
        if move.name is None:
            raise InvalidNameError
        move.user = self.address
        serialized = move.serialize(include_signature=False)
        move.signature = '{signature} {public_key}'.format(
            signature=seccure.sign(
                serialized,
                self.private_key.encode('utf-8')
            ).decode('utf-8'),
            public_key=self.public_key,
        )
        move.id = move.hash

    @property
    def moves(self):
        """ return the user's moves. """
        return self.session.query(Move).filter_by(user=self.address).filter(
            Move.block != None # noqa
        ).order_by(Move.block_id.desc())

    def move(self, new_move, tax=0, commit=True):
        """
        make a move of the user.

        :params new_move: Move object to make
        :params      tax: Tax to apply
        :params   commit: commit in this function automatically or not
        """
        new_move.user = self.address
        new_move.tax = tax
        new_move.created_at = datetime.datetime.utcnow()
        self.sign(new_move)

        if new_move.valid:
            if commit:
                self.session.add(new_move)
                self.session.commit()
        else:
            raise InvalidMoveError

        return new_move

    def hack_and_slash(self, weapon=None, armor=None, food=None):
        details = dict()
        if weapon:
            details['weapon'] = weapon
        if armor:
            details['armor'] = armor
        if food:
            details['food'] = food
        return self.move(HackAndSlash(details=details))

    def sleep(self, spot=''):
        return self.move(Sleep())

    def send(self, item_name, amount, receiver):
        return self.move(Send(details={
            'item_name': item_name,
            'amount': amount,
            'receiver': receiver,
        }))

    def sell(self, item_name, price):
        return self.move(Sell(details={'item_name': item_name,
                                       'price': price}))

    def buy(self, move_id):
        return self.move(Buy(details={'move_id': move_id}))

    def create_novice(self, details):
        return self.move(CreateNovice(details=details))

    def level_up(self, new_status):
        return self.move(LevelUp(details={
            'new_status': new_status,
        }))

    def say(self, content):
        return self.move(Say(details={'content': content}))

    def combine(self, item1, item2, item3):
        return self.move(Combine(details={'item1': item1,
                                          'item2': item2,
                                          'item3': item3}))

    def create_block(self, moves, commit=True, click=None):
        """ Create a block. """
        for move in moves:
            if not move.valid:
                raise InvalidMoveError(move)
        block = Block()
        block.root_hash = h(
            ''.join(sorted((m.id for m in moves))).encode('utf-8')
        ).hexdigest()
        block.creator = self.address
        block.created_at = datetime.datetime.utcnow()

        prev_block = self.session.query(Block).order_by(
            Block.id.desc()
        ).first()
        if prev_block:
            block.id = prev_block.id + 1
            block.prev_hash = prev_block.hash
            block.difficulty = prev_block.difficulty
            difficulty_check_block = self.session.query(Block).get(
                max(1, block.id - 10)
            )
            avg_timedelta = (
                (block.created_at - difficulty_check_block.created_at) /
                (block.id - difficulty_check_block.id)
            )
            if click:
                click.echo(
                    f'avg: {avg_timedelta}, difficulty: {block.difficulty}'
                )
            if avg_timedelta <= datetime.timedelta(0, 5):
                block.difficulty = max(0, block.difficulty + 1)
            elif avg_timedelta > datetime.timedelta(0, 15):
                block.difficulty = max(0, block.difficulty - 1)
        else:
            #: Genesis block
            block.id = 1
            block.prev_hash = None
            block.difficulty = 0

        block.suffix = hashcash._mint(
            block.serialize().decode('utf-8'),
            bits=block.difficulty
        )
        if self.session.query(Block).get(block.id):
            return None
        block.hash = h(
            (block.serialize().decode('utf-8') + block.suffix).encode('utf-8')
        ).hexdigest()

        for move in moves:
            move.block = block

        if not block.valid:
            raise InvalidBlockError

        if commit:
            self.session.add(block)
            self.session.commit()

        return block

    def avatar(self, block_id=None):
        """
        get avatar of the user.

        :params block_id: Avatar's block timing. If None, it sets last block.
        """
        if not block_id:
            block = self.session.query(Block).order_by(
                Block.id.desc()).first()
            if block:
                block_id = block.id
            else:
                block_id = 0
        return Avatar.get(self.address, block_id)


class Avatar():
    @classmethod
    @cache.memoize(timeout=0)
    def get(cls, user_addr, block_id, session=db.session):
        """
        get avatar.

        :params user_addr: Avatar's user address
        :params  block_id: Avatar's block timing
        :params   session: Database session to get data
        """
        create_move = session.query(Move).filter_by(user=user_addr).filter(
            Move.block_id <= block_id
        ).order_by(
            Move.block_id.desc()
        ).filter(
            Move.name.like('create_%')
        ).first()
        if not create_move or block_id < create_move.block_id:
            return None
        moves = session.query(Move).filter(
            or_(Move.user == user_addr, Move.id.in_(
                    db.session.query(MoveDetail.move_id).filter_by(
                        key='receiver', value=user_addr)))
        ).filter(
            Move.block_id >= create_move.block_id,
            Move.block_id <= block_id
        ).order_by(Move.block_id.asc())
        avatar, result = create_move.execute(None)
        avatar.items['GOLD'] += session.query(Block).filter_by(
            creator=user_addr
        ).filter(Block.id <= block_id).count() * 8

        for move in moves:
            if move.user == user_addr:
                avatar, result = move.execute(avatar)
            if (type(move) == Send and
               move.details['receiver'] == user_addr):
                avatar, result = move.receive(avatar)

        return avatar

    def modifier(self, status):
        """ Return modifier of the status. """
        status = getattr(self, status)
        if status in (1, 2, 3):
            return -3
        elif status in (4, 5):
            return -2
        elif status in (6, 7, 8):
            return -1
        elif status in (9, 10, 11, 12):
            return 0
        elif status in (13, 14, 15):
            return 1
        elif status in (16, 17):
            return 2
        elif status >= 18:
            return 3
        return 0

    def get_item(self, item):
        """
        Get given item.

        :params item: item's ticker name
        """
        if item not in self.items:
            self.items[item] = 1
        else:
            self.items[item] += 1

    @property
    def damage(self):
        raise NotImplementedError

    @property
    def max_hp(self):
        raise NotImplementedError

    @property
    def profile_image_url(self):
        return f'https://www.gravatar.com/avatar/{self.gravatar_hash}?d=mm'

    @property
    def weapons(self) -> list:
        result = []
        for weapon in get_related_items(Weapon):
            if (weapon.ticker_name in self.items.keys() and
               self.items[weapon.ticker_name] > 0):
                result.append(weapon)

        return result

    @property
    def last_weapon(self) -> Weapon:
        last_has = HackAndSlash.query.filter_by(
            user=self.user
        ).order_by(HackAndSlash.block_id.desc()).first()

        if last_has:
            try:
                return last_has.details['weapon']
            except KeyError:
                return None
        else:
            return None

    @property
    def last_armor(self) -> Armor:
        last_has = HackAndSlash.query.filter_by(
            user=self.user
        ).order_by(HackAndSlash.block_id.desc()).first()

        if last_has:
            try:
                return last_has.details['armor']
            except KeyError:
                return None
        else:
            return None

    @property
    def armors(self) -> list:
        result = []
        for armor in get_related_items(Armor):
            if (armor.ticker_name in self.items.keys() and
               self.items[armor.ticker_name] > 0):
                result.append(armor)

        return result

    @property
    def foods(self) -> list:
        result = []
        for food in get_related_items(Food):
            if (food.ticker_name in self.items.keys() and
               self.items[food.ticker_name] > 0):
                result.append(food)

        return result


class Novice(Avatar):
    @property
    def damage(self):
        return '1d6'

    @property
    def max_hp(self):
        return self.constitution + 6



import os

from flask import Flask

from nekoyume.models import cache, db
from nekoyume.api import api
from nekoyume.game import babel, game
from nekoyume.tasks import celery


def make_celery(app):
    celery.name = app.import_name
    celery.conf.update(
        result_backend=app.config['CELERY_RESULT_BACKEND'],
        broker_url=app.config['CELERY_BROKER_URL'])
    celery.conf.update(app.config)
    TaskBase = celery.Task

    class ContextTask(TaskBase):
        abstract = True

        def __call__(self, *args, **kwargs):
            with app.app_context():
                return TaskBase.__call__(self, *args, **kwargs)

    celery.Task = ContextTask
    return celery


def create_app():
    app = Flask(__name__)
    app.config['SQLALCHEMY_DATABASE_URI'] = os.environ.get(
        'DATABASE_URL', 'sqlite:///yume.db')
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.config['SQLALCHEMY_TRACK_MODIFICATIONS'] = False

    try:
        with open('.secret_key', 'rb') as f:
            app.secret_key = f.read()
    except FileNotFoundError:
        app.secret_key = os.urandom(64)
        f = open('.secret_key', 'wb')
        f.write(app.secret_key)
        f.close()

    app.register_blueprint(api)
    app.register_blueprint(game)

    db.app = app
    db.init_app(app)

    babel.app = app
    babel.init_app(app)

    app.config.update(
        CELERY_BROKER_URL=os.environ.get(
            'REDIS_URL', 'redis://localhost:6379'),
        CELERY_RESULT_BACKEND=os.environ.get(
            'REDIS_URL', 'redis://localhost:6379'))
    return app


app = create_app()
cel = make_celery(app)
cache.init_app(app, config={'CACHE_TYPE': 'redis',
                            'CACHE_REDIS_URL': os.environ.get(
                                'REDIS_URL', 'redis://localhost:6379'
                            )})


def run():
    from gunicorn.app import wsgiapp
    wsgiapp.run()



import datetime

from flask import Blueprint, jsonify, request
import requests

from nekoyume.tasks import block_broadcast, move_broadcast
from nekoyume.models import db, Block, Node, Move, get_my_public_url


api = Blueprint('api', __name__, template_folder='templates')


@api.route('/ping')
def get_pong():
    return 'pong'


@api.route('/public_url')
def get_public_url():
    return jsonify(
        url=get_my_public_url()
    )


@api.route(Node.get_nodes_endpoint, methods=['GET'])
def get_nodes():
    nodes = Node.query.filter(
        Node.last_connected_at >= datetime.datetime.now() -
        datetime.timedelta(60 * 3)
    ).order_by(Node.last_connected_at.desc()).limit(2500).all()

    public_url = get_my_public_url()
    if public_url:
        nodes.append(Node(url=public_url,
                          last_connected_at=datetime.datetime.now()))

    return jsonify(nodes=[n.url for n in nodes])


@api.route(Node.post_node_endpoint, methods=['POST'])
def post_node():
    if 'url' in request.values:
        url = request.values['url']
    elif 'url' in request.get_json():
        url = request.get_json()['url']
    else:
        return jsonify(
            result='failed',
            message='Invalid parameter.'
        ), 400
    node = Node.query.get(url)
    if node:
        return jsonify(result='success')
    else:
        node = Node(url=url)
        db.session.add(node)
        try:
            response = requests.get(f'{node.url}/ping')
        except requests.exceptions.ConnectionError:
            return jsonify(
                result='failed',
                message=f'Connection to node {node.url} was failed.'
            ), 403
        if response.status_code == 200:
            node.last_connected_at = datetime.datetime.now()
            db.session.commit()
            return jsonify(result='success')
        else:
            return jsonify(
                result='failed',
                message=f'Connection to node {node.url} was failed.'
            ), 403


@api.route(Node.get_blocks_endpoint, methods=['GET'])
def get_blocks():
    last_block = Block.query.order_by(Block.id.desc()).first()
    from_ = request.values.get('from', 1, type=int)
    to = request.values.get(
        'to',
        last_block.id if last_block else 0,
        type=int)
    blocks = Block.query.filter(
        Block.id >= from_,
        Block.id <= to
    ).order_by(Block.id.asc())
    return jsonify(blocks=[b.serialize(use_bencode=False,
                                       include_suffix=True,
                                       include_moves=True,
                                       include_hash=True)
                           for b in blocks])


@api.route('/blocks/<string:block_hash>')
def get_block_by_hash(block_hash):
    block = Block.query.filter_by(hash=block_hash).first()
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/blocks/<int:block_id>')
def get_block_by_id(block_id):
    block = Block.query.get(block_id)
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/blocks/last')
def get_last_block():
    block = Block.query.order_by(Block.id.desc()).first()
    if block:
        block = block.serialize(use_bencode=False,
                                include_suffix=True,
                                include_moves=True,
                                include_hash=True)
    return jsonify(block=block)


@api.route('/moves/<string:move_id>')
def get_moves(move_id):
    move = Move.query.get(move_id)
    if move:
        move = move.serialize(False, True, True, True)
    return jsonify(move=move)


@api.route(Node.post_block_endpoint, methods=['POST'])
def post_block():
    new_block = request.get_json()
    last_block = Block.query.order_by(Block.id.desc()).first()

    if not new_block:
        return jsonify(result='failed',
                       message="empty block."), 400

    if not last_block and new_block['id'] != 1:
        Block.sync(
            Node.query.order_by(
                Node.last_connected_at.desc()).first())
        return jsonify(result='failed',
                       message="new block isn't our next block."), 403

    if (new_block['id'] > 1 and
       (new_block['id'] != last_block.id + 1 or
       new_block['prev_hash'] != last_block.hash)):
        if new_block['id'] > last_block.id + 1:
            Block.sync(
                Node.query.order_by(
                    Node.last_connected_at.desc()).first())
        return jsonify(result='failed',
                       message="new block isn't our next block."), 403

    block = Block()
    block.id = new_block['id']
    block.creator = new_block['creator']
    block.created_at = datetime.datetime.strptime(
        new_block['created_at'], '%Y-%m-%d %H:%M:%S.%f')
    block.prev_hash = new_block['prev_hash']
    block.hash = new_block['hash']
    block.difficulty = new_block['difficulty']
    block.suffix = new_block['suffix']
    block.root_hash = new_block['root_hash']

    for new_move in new_block['moves']:
        move = Move.query.get(new_move['id'])
        if not move:
            move = Move(
                id=new_move['id'],
                user=new_move['user'],
                name=new_move['name'],
                signature=new_move['signature'],
                tax=new_move['tax'],
                details=new_move['details'],
                created_at=datetime.datetime.strptime(
                    new_move['created_at'], '%Y-%m-%d %H:%M:%S.%f'
                ),
                block_id=block.id,
            )
        if not move.valid:
            return jsonify(result='failed',
                           message=f"move {move.id} isn't valid."), 400
        block.moves.append(move)
    if not block.valid:
        return jsonify(result='failed',
                       message="new block isn't valid."), 400

    db.session.add(block)
    db.session.commit()
    sent_node = Node()
    if 'sent_node' in new_block:
        sent_node.url = new_block['sent_node']
    block_broadcast.delay(
        block.id,
        sent_node_url=sent_node.url,
        my_node_url=f'{request.scheme}://{request.host}'
    )
    return jsonify(result='success')


@api.route(Node.post_move_endpoint, methods=['POST'])
def post_move():
    new_move = request.get_json()
    move = Move.query.get(new_move['id'])

    if move:
        return jsonify(result='success')

    if not move:
        move = Move(
            id=new_move['id'],
            user=new_move['user'],
            name=new_move['name'],
            signature=new_move['signature'],
            tax=new_move['tax'],
            details=new_move['details'],
            created_at=datetime.datetime.strptime(
                new_move['created_at'], '%Y-%m-%d %H:%M:%S.%f'
            ),
        )

    if not move.valid:
        return jsonify(result='failed',
                       message=f"move {move.id} isn't valid."), 400

    db.session.add(move)
    db.session.commit()

    sent_node = Node()
    if 'sent_node' in new_move:
        sent_node.url = new_move['sent_node']

    move_broadcast.delay(
        move.id,
        sent_node_url=sent_node.url,
        my_node_url=f'{request.scheme}://{request.host}'
    )
    return jsonify(result='success')



import click
import time

from ptpython.repl import embed

from nekoyume.models import Node, Block, Move, User, get_my_public_url
from nekoyume.app import app, db


@click.group()
def cli():
    pass


@click.command()
@click.option('--private-key',
              default='test',
              help='Private key of neko')
def neko(private_key):
    app.app_context().push()
    while True:
        Block.sync()
        block = User(private_key).create_block(
            [m
             for m in Move.query.filter_by(block=None).limit(20).all()
             if m.valid],
            click=click,
        )
        if block:
            block.broadcast()
            click.echo(block)


@click.command()
def shell():
    app.app_context().push()
    embed(globals(), locals())


@click.command()
@click.option('--seed',
              default=None,
              help='Seed node URL to connect')
def init(seed):
    click.echo('Creating database...')
    db.create_all()
    click.echo(f'Updating node... (seed: {seed})')
    if seed:
        Node.update(Node.get(url=seed))
    else:
        Node.update()
    click.echo('Syncing blocks...')
    Block.sync(click=click)


@click.command()
def sync():
    public_url = get_my_public_url()
    if public_url:
        click.echo(f"You have a public node url. ({public_url})")
        Node.broadcast(Node.post_node_endpoint, {'url': public_url})
    while True:
        try:
            prev_id = Block.query.order_by(Block.id.desc()).first().id
        except AttributeError:
            click.echo("You need to initialize. try `nekoyume init`.")
            break
        if not prev_id:
            click.echo("You need to initialize. try `nekoyume init`.")
            break
        Block.sync(click=click)
        if prev_id == Block.query.order_by(Block.id.desc()).first().id:
            click.echo("The blockchain is up to date.")
            time.sleep(15)


cli.add_command(init)
cli.add_command(neko)
cli.add_command(shell)
cli.add_command(sync)


if __name__ == '__main__':
    cli()



from setuptools import setup, find_packages
from codecs import open
from os import path


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='nekoyume',
    version='0.0.2',
    description='Decentralized MMORPG based on Dungeon World',
    long_description=long_description,
    url='https://github.com/nekoyume/nekoyume',
    license='LICENSE.txt',
    author='JC Kim',
    author_email='jc@nekoyu.me',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Other Audience',
        'Topic :: Database :: Database Engines/Servers',
        'Topic :: Games/Entertainment :: Role-Playing',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
    ],
    keywords='blockchain mmorpg game',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=[
        'bencode.py >= 2.0.0, < 2.1.0',
        'celery >= 4.1.0, < 4.2.0',
        'click >= 6.7, < 7.0',
        'Flask >= 0.12.2, < 0.13.0',
        'Flask-Babel >= 0.11.2, < 0.12.0',
        'Flask-Cache >= 0.13.1, < 0.14.0',
        'Flask-SQLAlchemy >= 2.3.2, < 2.4.0',
        'gevent == 1.3.1',
        'gunicorn >= 19.7.1, < 19.8.0',
        'ptpython == 0.41',
        'pycrypto == 2.6',
        'python_bitcoinlib == 0.9.0',
        'pytz >= 2018.3',
        'redis >= 2.10.6, < 2.11.0',
        'requests >= 2.18.4, < 2.19.0',
        'seccure >= 0.3.2, < 0.4.0',
        'SQLAlchemy >= 1.2.2, < 1.3.0',
        'tablib >= 0.12.1, < 0.13.0',
    ],
    extras_require={
        'dev': [
            'flake8 >= 3.5.0, < 3.6.0',
            'recommonmark==0.4.0',
            'Sphinx >= 1.7.1, < 1.8.0',
            'sphinx-rtd-theme==0.2.4',
        ],
        'test': [
            'pytest >= 3.4.2, < 3.5.0',
            'pytest-localserver >= 0.4.1, < 0.5',
            'codecov >= 2.0.15, < 2.1.0',
        ],
    },
    package_data={
        'nekoyume': ['data/*', 'templates/*.html'],
    },
    entry_points={
        'console_scripts': [
            'nekoyume = nekoyume.cli:cli',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/nekoyume/nekoyume/issues',
        'Funding': 'https://nekoyu.me/',
        'Source': 'https://github.com/nekoyume/nekoyume/',
    },
)



from celery import Celery

from nekoyume.models import Block, db, Move, Node


celery = Celery()


@celery.task()
def move_broadcast(move_id, sent_node_url, my_node_url, session=db.session):
    try:
        session.query(Move).get(move_id).broadcast(Node(url=sent_node_url),
                                                   Node(url=my_node_url))
    except AttributeError:
        pass


@celery.task()
def block_broadcast(block_id, sent_node_url, my_node_url, session=db.session):
    try:
        session.query(Block).get(block_id).broadcast(Node(url=sent_node_url),
                                                     Node(url=my_node_url))
    except AttributeError:
        pass



from functools import wraps

from flask import (Blueprint, g, request, redirect, render_template,
                   session, url_for)
from flask_babel import Babel
from sqlalchemy import func

from nekoyume.models import cache, db, LevelUp, Move, Node, User


game = Blueprint('game', __name__, template_folder='templates')
babel = Babel()


@babel.localeselector
def get_locale():
    return request.accept_languages.best_match(['ko', 'en'])


def login_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        if session.get('private_key') is None:
            return redirect(url_for('.get_login', next=request.url))
        else:
            g.user = User(session['private_key'])
        return f(*args, **kwargs)
    return decorated_function


@game.route('/login', methods=['GET'])
def get_login():
    return render_template('login.html')


@game.route('/login', methods=['POST'])
def post_login():
    session['private_key'] = request.values.get('private_key')
    if 'next' in request.values:
        return redirect(request.values.get('next'))
    else:
        return redirect(url_for('.get_dashboard'))


@game.route('/logout', methods=['GET'])
@login_required
def get_logout():
    del session['private_key']
    return redirect(url_for('game.get_login'))


@cache.memoize(60)
def get_rank():
    return db.session.query(
        LevelUp.user, func.count(LevelUp.id)
    ).group_by(LevelUp.user).order_by(
        func.count(LevelUp.id).desc()
    ).limit(10).all()


def get_unconfirmed_move(address):
    unconfirmed_moves = Move.query.filter_by(
        user=address, block=None
    )
    for unconfirmed_move in unconfirmed_moves:
        if unconfirmed_move.valid:
            return unconfirmed_move
    return None


@game.route('/')
@login_required
def get_dashboard():
    if not g.user.avatar():
        return redirect(url_for('.get_new_novice'))

    unconfirmed_move = get_unconfirmed_move(g.user.address)

    feed = g.user.moves
    # for caching
    for move in reversed(feed.limit(10).all()):
        avatar, result = move.execute()
    return render_template('dashboard.html',
                           unconfirmed_move=unconfirmed_move,
                           feed=feed.order_by(Move.block_id.desc()),
                           rank=get_rank())


@game.route('/new')
@login_required
def get_new_novice():
    if not g.user.avatar():
        move = Move.query.filter_by(
            user=g.user.address,
            name='create_novice',
        ).first()
        if not move:
            move = g.user.create_novice({
                'strength': '12',
                'dexterity': '12',
                'constitution': '9',
                'intelligence': '10',
                'wisdom': '8',
                'charisma': '13'})
            db.session.add(move)
            db.session.commit()
            move.broadcast(
                my_node=Node(url=f'{request.scheme}://{request.host}')
            )
        return render_template('new.html', move=move)
    return redirect(url_for('.get_dashboard'))


@game.route('/session_moves', methods=['POST'])
@login_required
def post_move():
    unconfirmed_move = get_unconfirmed_move(g.user.address)

    if unconfirmed_move:
        return redirect(url_for('.get_dashboard'))

    if request.values.get('name') == 'hack_and_slash':
        move = g.user.hack_and_slash(request.values.get('weapon'),
                                     request.values.get('armor'),
                                     request.values.get('food'),)
    if request.values.get('name') == 'sleep':
        move = g.user.sleep()
    if request.values.get('name') == 'level_up':
        move = g.user.level_up(request.values.get('new_status'))
    if request.values.get('name') == 'say':
        move = g.user.say(request.values.get('content'))
    if request.values.get('name') == 'send':
        move = g.user.send(request.values.get('item'),
                           request.values.get('amount'),
                           request.values.get('receiver'))
    if request.values.get('name') == 'combine':
        move = g.user.combine(request.values.get('item1'),
                              request.values.get('item2'),
                              request.values.get('item3'))

    if move:
        move.broadcast(my_node=Node(url=f'{request.scheme}://{request.host}'))
    return redirect(url_for('.get_dashboard'))
